# THIS FILE IS AUTO-GENERATED. PLEASE DO NOT MODIFY
# version file for ldap3
# generated on 2021-07-18 08:33:22.864195
# on system uname_result(system='Windows', node='ELITE10GC', release='10', version='10.0.19041', machine='AMD64')
# with Python 3.9.4 - ('tags/v3.9.4:1f2e308', 'Apr  6 2021 13:40:21') - MSC v.1928 64 bit (AMD64)
#
__version__ = '0.0.5'
__author__ = 'Leo'
__email__ = 'yshan20@su.edu'
__url__ = 'https://github.com/462548187/google-ldap3.git'
__description__ = 'A strictly RFC 4510 conforming LDAP V3 pure Python client library'
__license__ = 'LGPL v3'
